# Generated by Django 3.0.3 on 2021-03-24 15:28
import django
from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0010_rolemutation'),
    ]

    operations = [
        migrations.DeleteModel(
            name='RoleMutation'
        ),
        migrations.CreateModel(
            name='RoleMutation',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False),),
                ('mutation', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='roles',
                                               to='core.MutationLog')),
                ('role', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='mutations',
                                           to='core.Role')),
            ],
            options={
                'db_table': 'core_RoleMutation',
                'managed': True,
            },
        ),
    ]
