# Generated by Django 3.0.14 on 2021-06-15 16:23

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0013_users_api'),
    ]

    operations = [
        migrations.AlterField(
            model_name='usermutation',
            name='core_user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='mutations', to=settings.AUTH_USER_MODEL),
        ),
    ]
